@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2016 PLT Design Inc.
+  Copyright (c) 2004-2017 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -262,6 +262,13 @@ Scheme_Comp_Env *scheme_new_compilation_frame(int num_bindings, int flags, Schem
 
   init_compile_data(frame);
 
+  if (flags & SCHEME_USE_SCOPES_TO_NEXT) {
+    if (base->use_scopes_next)
+      frame->use_scopes_next = base->use_scopes_next;
+    else
+      frame->use_scopes_next = base;
+  }
+
   return frame;
 }
 
@@ -978,7 +985,7 @@ void scheme_set_compilation_variables(Scheme_Comp_Env *frame, Scheme_IR_Local **
 
   for (i = 0; i < count; i++) {
     MZ_ASSERT(!frame->vars[i+pos]);
-    frame->vars[i+pos] = vars[i];
+    frame->vars[i+pos] = vars[count - i - 1];
   }
 }
 
@@ -1100,7 +1107,7 @@ void create_skip_table(Scheme_Comp_Env *start_frame)
   }
 
   if (!table) {
-    table = scheme_make_hash_tree(0);
+    table = scheme_make_hash_tree(SCHEME_hashtr_eq);
     table = scheme_hash_tree_set(table, scheme_make_integer(0), (Scheme_Object *)end_frame);
   }
 
@@ -1144,6 +1151,13 @@ Scheme_Object *scheme_intern_struct_proc_shape(int shape)
   return scheme_intern_symbol(buf);
 }
 
+Scheme_Object *scheme_intern_struct_prop_proc_shape(int shape)
+{
+  char buf[20];
+  sprintf(buf, "prop%d", shape);
+  return scheme_intern_symbol(buf);
+}
+
 void scheme_dump_env(Scheme_Comp_Env *env)
 {
   Scheme_Comp_Env *frame;
@@ -1588,6 +1602,11 @@ scheme_compile_lookup(Scheme_Object *find_id, Scheme_Comp_Env *env, int flags,
         *_inline_variant = mod_constant;
       is_constant = 2;
       shape = scheme_intern_struct_proc_shape(SCHEME_PROC_SHAPE_MODE(mod_constant));
+    } else if (SAME_TYPE(SCHEME_TYPE(mod_constant), scheme_struct_prop_proc_shape_type)) {
+      if (_inline_variant)
+        *_inline_variant = mod_constant;
+      is_constant = 2;
+      shape = scheme_intern_struct_prop_proc_shape(SCHEME_PROC_SHAPE_MODE(mod_constant));
     } else if (SAME_TYPE(SCHEME_TYPE(mod_constant), scheme_inline_variant_type)) {
       if (_inline_variant) {
         /* In case the inline variant includes references to module
