@@ -1,6 +1,6 @@
 /*
   Racket
-  Copyright (c) 2004-2016 PLT Design Inc.
+  Copyright (c) 2004-2017 PLT Design Inc.
   Copyright (c) 1995-2001 Matthew Flatt
 
     This library is free software; you can redistribute it and/or
@@ -128,6 +128,31 @@ uintptr_t PTR_TO_LONG(Scheme_Object *o)
 # define PTR_TO_LONG(p) ((uintptr_t)(p)>>2)
 #endif
 
+void scheme_install_symbol_hash_code(Scheme_Object *sym, uintptr_t h)
+{
+#ifdef MZ_PRECISE_GC
+  /* Record a hash code for the symbol as its `eq?` hash code ---
+     intended mainly to make `equal?` hashing depend only on the
+     symbol content */
+  short v;
+
+  v = sym->keyex;
+
+  if (!(v & 0xFFFC)) {
+    v |= (short)(h & ~0x7);
+#ifdef OBJHEAD_HAS_HASH_BITS
+    if (GC_is_allocated(sym)) {
+      OBJHEAD_HASH_BITS(sym) = (h >> 16);
+      v |= GCABLE_OBJ_HASH_BIT;
+    } else
+      v &= ~GCABLE_OBJ_HASH_BIT;
+#endif
+    if (!v) v = 0x1AD0;
+    sym->keyex = v;
+  }
+#endif
+}
+
 #define FILL_FACTOR 1.4
 
 #define MIN_HTABLE_SIZE 8
@@ -657,6 +682,37 @@ void scheme_reset_hash_table(Scheme_Hash_Table *table, int *history)
   table->mcount = 0;
 }
 
+Scheme_Object *scheme_hash_table_next(Scheme_Hash_Table *hash,
+				      mzlonglong start)
+{
+    int i, sz = hash->size;
+    if (start >= 0) {
+      if ((start >= sz) || !hash->vals[start])
+        return NULL;
+    }
+    for (i = start + 1; i < sz; i++) {
+      if (hash->vals[i])
+        return scheme_make_integer(i);
+    }
+
+    return scheme_false;
+}
+
+int scheme_hash_table_index(Scheme_Hash_Table *hash, mzlonglong pos,
+			    Scheme_Object **_key, Scheme_Object **_val)
+{
+  if (pos < hash->size) {
+    if (hash->vals[pos]) {
+      *_key = hash->keys[pos];
+      if (_val)
+	*_val = hash->vals[pos];
+      return 1;
+    }
+  }
+  
+  return 0;
+}
+
 /*========================================================================*/
 /*                  old-style hash table, with buckets                    */
 /*========================================================================*/
@@ -1103,6 +1159,47 @@ Scheme_Bucket_Table *scheme_clone_bucket_table(Scheme_Bucket_Table *bt)
   return table;
 }
 
+Scheme_Object *scheme_bucket_table_next(Scheme_Bucket_Table *hash,
+					mzlonglong start)
+{
+  Scheme_Bucket *bucket;
+  int i, sz = hash->size;
+    
+  if (start >= 0) {
+    bucket = ((start < sz) ? hash->buckets[start] : NULL);
+    if (!bucket || !bucket->val || !bucket->key) 
+      return NULL;      
+  }
+  for (i = start + 1; i < sz; i++) {
+    bucket = hash->buckets[i];
+    if (bucket && bucket->val && bucket->key) {
+      return scheme_make_integer(i);
+    }
+  }
+
+  return scheme_false;
+}
+
+int scheme_bucket_table_index(Scheme_Bucket_Table *hash, mzlonglong pos, Scheme_Object **_key, Scheme_Object **_val)
+{
+  Scheme_Bucket *bucket;
+
+  if (pos < hash->size) {
+    bucket = hash->buckets[pos];
+    if (bucket && bucket->val && bucket->key) {
+      if (hash->weak)
+	*_key = (Scheme_Object *)HT_EXTRACT_WEAK(bucket->key);
+      else
+	*_key = (Scheme_Object *)bucket->key;
+      if (_val)
+	*_val = (Scheme_Object *)bucket->val;
+      return 1;
+    }
+  }
+  
+  return 0;
+}
+
 /*========================================================================*/
 /*                         precise GC hashing                             */
 /*========================================================================*/
@@ -1138,6 +1235,11 @@ typedef struct Hash_Info {
 static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi);
 static uintptr_t equal_hash_key2(Scheme_Object *o, Hash_Info *hi);
 
+/* Based on Bob Jenkins's one-at-a-time hash function at
+   http://www.burtleburtle.net/bob/hash/doobs.html: */
+#define MZ_MIX(k) (k += (k << 10), k ^= (k >> 6))
+
+
 static Scheme_Object *hash_recur(int argc, Scheme_Object **argv, Scheme_Object *prim)
 {
   intptr_t v;
@@ -1194,94 +1296,71 @@ static uintptr_t overflow_equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Inf
 XFORM_NONGCING static uintptr_t dbl_hash_val(double d) 
   XFORM_SKIP_PROC
 {
-  int e;
-  
-  if (MZ_IS_NAN(d)) {
-    d = 0.0;
-    e = 1000;
-  } else if (MZ_IS_POS_INFINITY(d)) {
-    d = 0.5;
-    e = 1000;
-  } else if (MZ_IS_NEG_INFINITY(d)) {
-    d = -0.5;
-    e = 1000;
-  } else if (!d && scheme_minus_zero_p(d)) {
-    d = 0;
-    e = 1000;
-  } else {
-    /* frexp should not be used on inf or nan: */
-    d = frexp(d, &e);
-  }
+  const umzlonglong m = (umzlonglong)0x880355f21e6d1965;
+  umzlonglong h = (16 * m);
+  umzlonglong v;
 
-  return ((uintptr_t)(intptr_t)(d * (1 << 30))) + (uintptr_t)e;
+  if (MZ_IS_NAN(d))
+    return 0;
+
+  MZ_ASSERT(sizeof(d) == sizeof(v));
+
+  memcpy(&v, &d, sizeof(umzlonglong));
+
+  /* based one https://code.google.com/archive/p/fast-hash/ */
+#define MZ_MIX_FOR_DBL(h) \
+  (h) ^= (h) >> 23;       \
+  (h) *= (umzlonglong)0x2127599bf4325c37; \
+  (h) ^= (h) >> 47;
+
+  MZ_MIX_FOR_DBL(v);
+  h ^= v;
+  h *= m;
+
+  MZ_MIX_FOR_DBL(h);
+  
+  return (uintptr_t)h;
 }
 
 XFORM_NONGCING static uintptr_t dbl_hash2_val(double d)  
   XFORM_SKIP_PROC
 {
-  int e;
-  
-  if (MZ_IS_NAN(d)
-      || MZ_IS_POS_INFINITY(d)
-      || MZ_IS_NEG_INFINITY(d)) {
-    e = 1;
-  } else {
-    /* frexp should not be used on inf or nan: */
-    d = frexp(d, &e);
-  }
-  return to_unsigned_hash(e);
+  return dbl_hash_val(d) >> (sizeof(uintptr_t) / 2);
 }
 
 #ifdef MZ_LONG_DOUBLE
 XFORM_NONGCING static uintptr_t long_dbl_hash_val(long_double d) 
   XFORM_SKIP_PROC
 {
-  int e;
-  
-  if (MZ_IS_LONG_NAN(d)) {
-    d = get_long_double_zero();
-    e = 1000;
-  } else if (MZ_IS_LONG_POS_INFINITY(d)) {
-    d = get_long_double_one_half();
-    e = 1000;
-  } else if (MZ_IS_LONG_NEG_INFINITY(d)) {
-    d = long_double_neg(get_long_double_one_half());
-    e = 1000;
-  } else if (long_double_eqv(d, get_long_double_zero()) && scheme_long_minus_zero_p(d)) {
-    d = get_long_double_zero();
-    e = 1000;
-  } else {
-    /* frexpl should not be used on inf or nan: */
-    d = long_double_frexp(d, &e);
+  char s[LONG_DOUBLE_BYTE_LEN];
+  int i;
+  uintptr_t k;
+
+  if (MZ_IS_LONG_NAN(d))
+    return 0;
+
+  /* Like `extfl->floating-point-bytes`, we assume that the
+     content of a `long double' occupies the first 10 bytes: */
+  memcpy(s, &d, LONG_DOUBLE_BYTE_LEN);
+
+  k = 0;
+  for (i = LONG_DOUBLE_BYTE_LEN; i--; ) {
+    k += s[i];
+    MZ_MIX(k);
   }
 
-  return uintptr_from_long_double(long_double_mult_i(d, 1<<30)) + e;
-  /*return ((uintptr_t)(d * (1 << 30))) + e;*/
+  return k;  
 }
 
 XFORM_NONGCING static uintptr_t long_dbl_hash2_val(long_double d)  
   XFORM_SKIP_PROC
 {
-  int e;
-  
-  if (MZ_IS_LONG_NAN(d)
-      || MZ_IS_LONG_POS_INFINITY(d)
-      || MZ_IS_LONG_NEG_INFINITY(d)) {
-    e = 1;
-  } else {
-    /* frexp should not be used on inf or nan: */
-    d = long_double_frexp(d, &e);
-  }
-  return to_unsigned_hash(e);
+  return long_dbl_hash_val(d) >> (sizeof(uintptr_t) / 2);
 }
 #endif
 
 #define OVERFLOW_HASH() overflow_equal_hash_key(o, k - t, hi)
 
-/* Based on Bob Jenkins's one-at-a-time hash function at
-   http://www.burtleburtle.net/bob/hash/doobs.html: */
-#define MZ_MIX(k) (k += (k << 10), k ^= (k >> 6))
-
 XFORM_NONGCING static uintptr_t fast_equal_hash_key(Scheme_Object *o, uintptr_t k, int *_done)
 /* must cover eqv hash keys that are just eq hash keys */
 {
@@ -1289,9 +1368,14 @@ XFORM_NONGCING static uintptr_t fast_equal_hash_key(Scheme_Object *o, uintptr_t
 
  top:
   t = SCHEME_TYPE(o);
-  k += t;
   
   switch(t) {
+  case scheme_true_type:
+    return k + PTR_TO_LONG(o);
+    break;
+  case scheme_false_type:
+    return k + PTR_TO_LONG(o);
+    break;
   case scheme_integer_type:
     {
       uintptr_t iv = to_unsigned_hash(SCHEME_INT_VAL(o));
@@ -1387,9 +1471,10 @@ XFORM_NONGCING static uintptr_t fast_equal_hash_key(Scheme_Object *o, uintptr_t
       if (!(MZ_OPT_HASH_KEY(&s->iso) & 0x1)) {
 	/* Interned. Make key depend only on the content. */
 	if (!(MZ_OPT_HASH_KEY(&s->iso) & 0xFFFC)) {
-	  int i, h = 0;
+	  int i;
+          unsigned int h = 0;
 	  for (i = s->len; i--; ) {
-	    h += (h << 5) + h + s->s[i];
+	    h += (h << 5) + h + (unsigned int)s->s[i];
 	  }
 	  h += (h << 2);
 	  if (!(((short)h) & 0xFFFC))
@@ -1401,6 +1486,10 @@ XFORM_NONGCING static uintptr_t fast_equal_hash_key(Scheme_Object *o, uintptr_t
       } else
 	return k + PTR_TO_LONG(o);
     }
+# else
+  case scheme_keyword_type:
+  case scheme_symbol_type:
+    return k + PTR_TO_LONG(o);
 # endif
   default:
     {
@@ -1597,6 +1686,12 @@ static uintptr_t equal_hash_key(Scheme_Object *o, uintptr_t k, Hash_Info *hi)
 	
           hi->depth += 2;
 
+          /* Use structure-type names (especially for prefabs): */
+          for (i = s1->stype->name_pos+1; i--; ) {
+            k += equal_hash_key(s1->stype->parent_types[i]->name, 0, hi);
+            MZ_MIX(k);
+          }
+
           for (i = SCHEME_STRUCT_NUM_SLOTS(s1); i--; ) {
             if (SAME_OBJ(o, orig_obj))
               elem = s1->slots[i];
@@ -1799,7 +1894,7 @@ intptr_t scheme_equal_hash_key(Scheme_Object *o)
   uintptr_t k;
   int done = 1;
 
-  k = fast_equal_hash_key(o, 0, &done);
+  k = fast_equal_hash_key(o, SCHEME_TYPE(o), &done);
   if (done)
     return to_signed_hash(k);
   else
@@ -1827,6 +1922,10 @@ XFORM_NONGCING static uintptr_t fast_equal_hash_key2(Scheme_Object *o, int *_don
   *_done = 1;
 
   switch(t) {
+  case scheme_true_type:
+    return 1;
+  case scheme_false_type:
+    return 2;
   case scheme_integer_type:
     return t - SCHEME_INT_VAL(o);
 #ifdef MZ_USE_SINGLE_FLOATS
@@ -2322,6 +2421,10 @@ intptr_t scheme_eqv_hash_key2(Scheme_Object *o)
 #define HASHTR_HAS_VAL  0x1
 #define HASHTR_HAS_CODE 0x2
 
+/* In a hash tree without HASHTR_HAS_VAL, all values are `#t`; we nodes without
+  HASHTR_HAS_VAL to nodes with it on demand, but we don't go the other way */
+#define NOT_IMPLICIT_VALUE(v) ((v) && !SAME_OBJ(v, scheme_true))
+
 #define HASHTR_SUBTREEP(o) SAME_TYPE(SCHEME_TYPE(o), scheme_hash_tree_subtree_type)
 #define HASHTR_COLLISIONP(o) SAME_TYPE(SCHEME_TYPE(o), scheme_hash_tree_collision_type)
 
@@ -2329,6 +2432,15 @@ intptr_t scheme_eqv_hash_key2(Scheme_Object *o)
 #define mzHAMT_LOG_WORD_SIZE 5
 #define mzHAMT_WORD_SIZE (1 << mzHAMT_LOG_WORD_SIZE)
 
+/* Reorder bits to make the HAMT more bushy in the case that sequential
+   hash code are added: */
+#define HAMT_REORDER(h) (((h) & (HIGH_PART | LOW_LOW_PARTS | (LOW_LOW_PARTS << PART_BIT_COUNT))) \
+                         | (((h) & HIGH_LOW_PARTS) << PART_BIT_COUNT)    \
+                         | (((h) & (HIGH_LOW_PARTS << PART_BIT_COUNT)) >> PART_BIT_COUNT))
+#define HIGH_PART (~((uintptr_t)0x0) - (uintptr_t)0xFFFFFFFF)
+#define LOW_LOW_PARTS ((uintptr_t)0x3333)
+#define HIGH_LOW_PARTS ((uintptr_t)0xCCCC)
+#define PART_BIT_COUNT 16
 
 XFORM_NONGCING static Scheme_Hash_Tree *resolve_placeholder(Scheme_Hash_Tree *ht)
 {
@@ -2345,7 +2457,7 @@ XFORM_NONGCING static int hamt_index(uintptr_t code, int shift)
   return (code >> shift) & ((1 << mzHAMT_LOG_WORD_SIZE) - 1);
 }
 
-XFORM_NONGCING int hamt_popcount(hash_tree_bitmap_t x)
+XFORM_NONGCING static int hamt_popcount(hash_tree_bitmap_t x)
 {
 #if MZ_HAS_BUILTIN_POPCOUNT
   return __builtin_popcount(x);
@@ -2376,7 +2488,7 @@ XFORM_NONGCING static hash_tree_bitmap_t hamt_bit(int index)
 
 XFORM_NONGCING Scheme_Object *_mzHAMT_VAL(Scheme_Hash_Tree *ht, int pos, int popcount)
 {
-  return ((SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL) ? (ht)->els[(popcount)+(pos)] : (ht)->els[pos]);
+  return ((SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL) ? (ht)->els[(popcount)+(pos)] : scheme_true);
 }
 
 XFORM_NONGCING uintptr_t mzHAMT_KEY_CODE(Scheme_Object *o)
@@ -2384,8 +2496,11 @@ XFORM_NONGCING uintptr_t mzHAMT_KEY_CODE(Scheme_Object *o)
   while (1) {
     if (HASHTR_COLLISIONP(o))
       o = ((Scheme_Hash_Tree *)o)->els[0];
-    else
-      return PTR_TO_LONG(o);
+    else {
+      uintptr_t h;
+      h = PTR_TO_LONG(o);
+      return HAMT_REORDER(h);
+    }
   }
 }
 
@@ -2411,6 +2526,11 @@ XFORM_NONGCING static void hamt_content_copy(Scheme_Hash_Tree *dest, Scheme_Hash
     if (SCHEME_HASHTR_FLAGS(src) & HASHTR_HAS_CODE) {
       memcpy(dest->els+2*dest_popcount+dest_start, src->els+2*src_popcount+src_start, len*sizeof(Scheme_Object*));
     }
+  } else if (SCHEME_HASHTR_FLAGS(dest) & HASHTR_HAS_VAL) {
+    /* make source's implicit `#t` values explicit in dest */
+    while (len--) {
+      _mzHAMT_SET_VAL(dest, dest_start+len, scheme_true, dest_popcount);
+    }
   }
 }
 
@@ -2445,14 +2565,23 @@ static Scheme_Hash_Tree *hamt_alloc(int kind, int popcount)
   return (Scheme_Hash_Tree *)scheme_malloc_small_tagged(HASH_TREE_RECORD_SIZE(kind, popcount));
 }
 
-static Scheme_Hash_Tree *hamt_dup(Scheme_Hash_Tree *ht, int popcount)
+static Scheme_Hash_Tree *hamt_dup(Scheme_Hash_Tree *ht, int popcount, int need_value)
 {
   Scheme_Hash_Tree *new_ht;
   int kind;
 
   kind = SCHEME_HASHTR_KIND(ht);
-  new_ht = hamt_alloc(kind, popcount);
+  new_ht = hamt_alloc(kind | (need_value ? HASHTR_HAS_VAL : 0), popcount);
   memcpy(new_ht, ht, HASH_TREE_RECORD_SIZE(kind, popcount));
+  
+  if (!(kind & HASHTR_HAS_VAL) && need_value) {
+    /* make original's implicit `#t` values explicit in the copy */
+    int i;
+    SCHEME_HASHTR_FLAGS(new_ht) |= HASHTR_HAS_VAL;
+    for (i = popcount; i--; ) {
+      _mzHAMT_SET_VAL(new_ht, i, scheme_true, popcount);
+    } 
+  }
 
   return new_ht;
 }
@@ -2493,6 +2622,7 @@ static Scheme_Hash_Tree *hamt_make2(int kind, int shift,
                         code2, key2, val2);
     return hamt_make1(new_ht, index1);
   } else {
+    kind = kind | ((NOT_IMPLICIT_VALUE(val1) || NOT_IMPLICIT_VALUE(val2)) ? HASHTR_HAS_VAL : 0);
     new_ht = hamt_alloc(kind, 2);
     new_ht->iso.so.type = scheme_hash_tree_subtree_type;
     SCHEME_HASHTR_FLAGS(new_ht) = kind;
@@ -2537,7 +2667,7 @@ static Scheme_Hash_Tree *hamt_set(Scheme_Hash_Tree *ht, uintptr_t code, int shif
   
   if (ht->bitmap & hamt_bit(index)) {
     /* Replacing: */
-    new_ht = hamt_dup(ht, popcount);
+    new_ht = hamt_dup(ht, popcount, NOT_IMPLICIT_VALUE(val));
     if (HASHTR_SUBTREEP(ht->els[pos])) {
       ht = (Scheme_Hash_Tree *)ht->els[pos];
       ht = hamt_set(ht, code, shift + mzHAMT_LOG_WORD_SIZE, key, val, inc);
@@ -2546,7 +2676,7 @@ static Scheme_Hash_Tree *hamt_set(Scheme_Hash_Tree *ht, uintptr_t code, int shif
     } else {
       if (code == _mzHAMT_CODE(new_ht, pos, popcount)) {
         new_ht->els[pos] = key;
-        if (SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL)
+        if (SCHEME_HASHTR_FLAGS(new_ht) & HASHTR_HAS_VAL)
           _mzHAMT_SET_VAL(new_ht, pos, val, popcount);
         new_ht->count += inc;
       } else {
@@ -2555,23 +2685,25 @@ static Scheme_Hash_Tree *hamt_set(Scheme_Hash_Tree *ht, uintptr_t code, int shif
                         _mzHAMT_CODE(new_ht, pos, popcount), new_ht->els[pos], _mzHAMT_VAL(new_ht, pos, popcount),
                         code, key, val);
         new_ht->els[pos] = (Scheme_Object *)ht;
-        if (SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL)
+        if (SCHEME_HASHTR_FLAGS(new_ht) & HASHTR_HAS_VAL)
           _mzHAMT_SET_VAL(new_ht, pos, NULL, popcount);
         new_ht->count += inc;
       }
     }
   } else {
-    new_ht = hamt_alloc(SCHEME_HASHTR_KIND(ht), popcount+1);
-    memcpy(new_ht, ht, HASH_TREE_RECORD_SIZE(SCHEME_HASHTR_KIND(new_ht), 0));
-    hamt_content_copy(new_ht, ht, popcount+1,popcount, 0, 0, pos);
+    int kind = SCHEME_HASHTR_KIND(ht) | (NOT_IMPLICIT_VALUE(val) ? HASHTR_HAS_VAL : 0);
+    new_ht = hamt_alloc(kind, popcount+1);
+    memcpy(new_ht, ht, HASH_TREE_RECORD_SIZE(SCHEME_HASHTR_KIND(ht), 0));
+    SCHEME_HASHTR_FLAGS(new_ht) |= kind;
+    hamt_content_copy(new_ht, ht, popcount+1, popcount, 0, 0, pos);
     if (pos < popcount)
       hamt_content_copy(new_ht, ht, popcount+1, popcount, pos+1, pos, popcount-pos);
     new_ht->bitmap |= hamt_bit(index);
     new_ht->count += inc;
     new_ht->els[pos] = key;
-    if (SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL) {
+    if (SCHEME_HASHTR_FLAGS(new_ht) & HASHTR_HAS_VAL) {
       _mzHAMT_SET_VAL(new_ht, pos, val, popcount+1);
-      if (SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_CODE)
+      if (SCHEME_HASHTR_FLAGS(new_ht) & HASHTR_HAS_CODE)
         _mzHAMT_SET_CODE(new_ht, pos, code, popcount+1);
     }
   }
@@ -2588,7 +2720,7 @@ static Scheme_Hash_Tree *hamt_contract(Scheme_Hash_Tree *ht, int popcount, int i
     return NULL;
 
   new_ht = hamt_alloc(SCHEME_HASHTR_KIND(ht), popcount-1);
-  memcpy(new_ht, ht, HASH_TREE_RECORD_SIZE(SCHEME_HASHTR_KIND(new_ht), 0));
+  memcpy(new_ht, ht, HASH_TREE_RECORD_SIZE(SCHEME_HASHTR_KIND(ht), 0));
   hamt_content_copy(new_ht, ht, popcount-1, popcount, 0, 0, pos);
   if (pos < popcount-1)
     hamt_content_copy(new_ht, ht, popcount-1, popcount, pos, pos+1, popcount-pos-1);
@@ -2615,11 +2747,15 @@ static Scheme_Hash_Tree *hamt_remove(Scheme_Hash_Tree *ht, uintptr_t code, int s
       if (!SAME_OBJ((Scheme_Object *)sub_ht, ht->els[pos])) {
         if (!sub_ht)
           return hamt_contract(ht, popcount, index, pos);
-        ht = hamt_dup(ht, popcount);
+        ht = hamt_dup(ht, popcount, 0);
         ht->count -= 1;
-        if ((sub_ht->count == 1) && !HASHTR_SUBTREEP(sub_ht->els[0])) {
+        if (((sub_ht->count == 1) && !HASHTR_SUBTREEP(sub_ht->els[0]))
+            || (HASHTR_COLLISIONP(sub_ht->els[0])
+                && (sub_ht->count == ((Scheme_Hash_Tree *)sub_ht->els[0])->count))) {
           /* drop extra layer that has 1 immediate entry */
           ht->els[pos] = sub_ht->els[0];
+          if (!(SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL) && (SCHEME_HASHTR_FLAGS(sub_ht) & HASHTR_HAS_VAL))
+            ht = hamt_dup(ht, popcount, 1);
           if (SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_VAL) {
             _mzHAMT_SET_VAL(ht, pos, _mzHAMT_VAL(sub_ht, 0, 1), popcount);
             if (SCHEME_HASHTR_FLAGS(ht) & HASHTR_HAS_CODE)
@@ -2635,17 +2771,24 @@ static Scheme_Hash_Tree *hamt_remove(Scheme_Hash_Tree *ht, uintptr_t code, int s
     return ht;
 }
 
+/* this signature is different from other scheme_<hash type>_next fns */
+/* but is used else where, so leave as is */
 mzlonglong scheme_hash_tree_next(Scheme_Hash_Tree *tree, mzlonglong pos)
 {
-  if (pos == -1)
-    pos = 0;
-  else
-    pos++;
-  
-  if (pos == tree->count)
+  mzlonglong i = pos+1;
+  if (i == tree->count)
     return -1;
   else
-    return pos;
+    return i;
+}
+
+Scheme_Object *scheme_hash_tree_next_pos(Scheme_Hash_Tree *tree, mzlonglong pos)
+{
+  mzlonglong i = pos+1;
+  if (i == tree->count)
+    return scheme_false;
+  else
+    return scheme_make_integer_value_from_long_long(i);
 }
 
 #if REVERSE_HASH_TABLE_ORDER
@@ -2656,6 +2799,195 @@ mzlonglong scheme_hash_tree_next(Scheme_Hash_Tree *tree, mzlonglong pos)
 # define HAMT_TRAVERSE_NEXT(i) ((i)+1)
 #endif
 
+#define mzHAMT_MAX_INDEX_LEVEL 4 /* For the compressed form of the index */
+
+/* instead of returning a pos, these unsafe iteration ops */
+/* return a view into the tree consisting of a: */
+/*   - subtree */
+/*   - subtree index */
+/*   - stack of parent subtrees and indices */
+/* For small hashes, it uses a compressed representation as fixnums */
+/*   - (#h i0) --> (+ (<< 1 5) i0) */
+/*   - (#h i1 #h i0) --> (+ (<< 1 10) (<< i1 5) i0) */
+/*   - (#h i2 #h i1 #h i0) --> (+ (<< 1 15) (<< i2 10) (<< i1 5) i0) */
+/*   - ...  */
+/* This speeds up performance of immutable hash table iteration. */
+/* These unsafe ops currently do not support REVERSE_HASH_TABLE_ORDER */
+/* to avoid unneeded popcount computations */
+Scheme_Object *scheme_unsafe_hash_tree_start(Scheme_Hash_Tree *ht)
+{
+  Scheme_Object *stack = scheme_null;
+  int j, i, i_n[mzHAMT_MAX_INDEX_LEVEL], level;
+  Scheme_Hash_Tree *ht_n[mzHAMT_MAX_INDEX_LEVEL];
+
+  ht = resolve_placeholder(ht);
+
+  if (!ht->count)
+    return scheme_false;
+
+  i = hamt_popcount(ht->bitmap)-1;
+  level = 0;
+  
+  while ((HASHTR_SUBTREEP(ht->els[i])
+          || HASHTR_COLLISIONP(ht->els[i]))) {
+    /* go down tree but save return point */
+    if (level == -1) {
+      stack = scheme_make_pair((Scheme_Object *)ht,
+                               scheme_make_pair(scheme_make_integer(i),
+                                                stack));
+    } else if (level < mzHAMT_MAX_INDEX_LEVEL) {
+      ht_n[level] = ht;
+      i_n[level] = i;
+      level++;
+    } else {
+      stack = scheme_null;
+      for (j = 0; j < mzHAMT_MAX_INDEX_LEVEL; j++) {
+        stack = scheme_make_pair((Scheme_Object *)ht_n[j],
+                                  scheme_make_pair(scheme_make_integer(i_n[j]),
+                                                   stack));
+      }
+      stack = scheme_make_pair((Scheme_Object *)ht,
+                               scheme_make_pair(scheme_make_integer(i),
+                                                stack));
+      level = -1;
+    }
+    ht = (Scheme_Hash_Tree *)ht->els[i];
+    i = hamt_popcount(ht->bitmap)-1;
+  }
+
+  if (level == -1) {
+    stack = scheme_make_pair((Scheme_Object *)ht,
+                             scheme_make_pair(scheme_make_integer(i),
+                                              stack));
+    return stack;
+  } else {
+    i = (1<<mzHAMT_LOG_WORD_SIZE) + i;
+    for (j = level-1; j >= 0 ; j--) {
+      i = (i<<mzHAMT_LOG_WORD_SIZE) + i_n[j];
+    }
+    return scheme_make_integer(i);
+  }
+}
+
+/* args is a (cons subtree (cons subtree-index stack-of-parents))
+   or the comppressed representation as a fixnum */
+XFORM_NONGCING void scheme_unsafe_hash_tree_subtree(Scheme_Object *obj, Scheme_Object *args,
+                                                    Scheme_Hash_Tree **_subtree, int *_i)
+{
+  Scheme_Hash_Tree *subtree;
+  int i;
+
+  if (SCHEME_PAIRP(args)) {
+    subtree = (Scheme_Hash_Tree *)SCHEME_CAR(args);
+    i = SCHEME_INT_VAL(SCHEME_CADR(args));
+  } else {
+    if (SCHEME_NP_CHAPERONEP(obj))
+      subtree = (Scheme_Hash_Tree *)SCHEME_CHAPERONE_VAL(obj);
+    else
+      subtree = (Scheme_Hash_Tree *)obj;
+    subtree = resolve_placeholder(subtree);
+    i = SCHEME_INT_VAL(args);
+    while (i >= (1<<(2*mzHAMT_LOG_WORD_SIZE))) {
+      subtree = (Scheme_Hash_Tree *)subtree->els[i & ((1<<mzHAMT_LOG_WORD_SIZE)-1)];
+      i = i >> mzHAMT_LOG_WORD_SIZE;
+    }
+    i = i & ((1<<mzHAMT_LOG_WORD_SIZE)-1);
+  }
+  
+  *_subtree = subtree;
+  *_i =i;
+}
+
+XFORM_NONGCING Scheme_Object *scheme_unsafe_hash_tree_access(Scheme_Hash_Tree *subtree, int i)
+{
+  return _mzHAMT_VAL(subtree, i, hamt_popcount(subtree->bitmap));
+}
+
+/* args is a (cons subtree (cons subtree-index stack-of-parents))
+   or the comppressed representation as a fixnum */
+Scheme_Object *scheme_unsafe_hash_tree_next(Scheme_Hash_Tree *ht, Scheme_Object *args)
+{
+  Scheme_Object *stack = scheme_null;
+  int j, i, i_n[mzHAMT_MAX_INDEX_LEVEL], level;
+  Scheme_Hash_Tree *ht_n[mzHAMT_MAX_INDEX_LEVEL];
+
+  if (SCHEME_PAIRP(args)) {
+    ht = (Scheme_Hash_Tree *)SCHEME_CAR(args);
+    i = SCHEME_INT_VAL(SCHEME_CADR(args));
+    stack = SCHEME_CDDR(args);
+    level = -1; /* -1 = too big */
+  } else {
+    ht = resolve_placeholder(ht);
+    i = SCHEME_INT_VAL(args);
+    level = 0;
+    while (i >= (1<<(2*mzHAMT_LOG_WORD_SIZE))) {
+      ht_n[level] = ht;
+      i_n[level] = i & ((1<<mzHAMT_LOG_WORD_SIZE)-1);
+      ht = (Scheme_Hash_Tree *)ht->els[i_n[level]];
+      i = i >> mzHAMT_LOG_WORD_SIZE;
+      level++;
+    }
+    i = i & ((1<<mzHAMT_LOG_WORD_SIZE)-1);
+  }
+
+  while (1) {
+    if (!i) { /* pop up the tree */
+      if (level == -1) {
+        ht = (Scheme_Hash_Tree *)SCHEME_CAR(stack);
+        i = SCHEME_INT_VAL(SCHEME_CADR(stack));
+        stack = SCHEME_CDDR(stack);
+        if (SCHEME_NULLP(stack))
+          level = 0;
+      } else if (!level) {
+        return scheme_false;
+      } else {
+        level--;
+        ht = ht_n[level];
+        i = i_n[level];
+      }
+    } else { /* go to next node */
+      i--;
+      if (!(HASHTR_SUBTREEP(ht->els[i])
+            || HASHTR_COLLISIONP(ht->els[i]))) {
+        if (level == -1) {
+          stack = scheme_make_pair((Scheme_Object *)ht,
+                                   scheme_make_pair(scheme_make_integer(i),
+                                                    stack));
+          return stack;
+        } else {
+          i = (1<<mzHAMT_LOG_WORD_SIZE) + i;
+          for (j = level-1; j >= 0 ; j--) {
+            i = (i<<mzHAMT_LOG_WORD_SIZE) + i_n[j];
+          }
+          return scheme_make_integer(i);
+        }
+      } else { /* go down tree but save return point */
+        if (level == -1) {
+          stack = scheme_make_pair((Scheme_Object *)ht,
+                                   scheme_make_pair(scheme_make_integer(i),
+                                                    stack));
+        } else if (level < mzHAMT_MAX_INDEX_LEVEL) {
+          ht_n[level] = ht;
+          i_n[level] = i;
+          level++;
+        } else {
+          stack = scheme_null;
+          for (j = 0; j < mzHAMT_MAX_INDEX_LEVEL; j++) {
+            stack = scheme_make_pair((Scheme_Object *)ht_n[j],
+                                     scheme_make_pair(scheme_make_integer(i_n[j]),
+                                                      stack));
+          }
+          stack = scheme_make_pair((Scheme_Object *)ht,
+                                   scheme_make_pair(scheme_make_integer(i),
+                                                    stack));
+          level = -1;
+        }
+        ht = (Scheme_Hash_Tree *)ht->els[i];
+        i = hamt_popcount(ht->bitmap);
+      }
+    }
+  }
+}
 
 XFORM_NONGCING static void hamt_at_index(Scheme_Hash_Tree *ht, mzlonglong pos,
                                          Scheme_Object **_key, Scheme_Object **_val, uintptr_t *_code)
@@ -2759,7 +3091,7 @@ XFORM_NONGCING static uintptr_t hamt_find_free_code(Scheme_Hash_Tree *tree, int
   Scheme_Hash_Tree *subtree;
   
   for (i = 0; i < mzHAMT_WORD_SIZE; i++) {
-    if (!(tree->bitmap & (1 << i)))
+    if (!(tree->bitmap & ((unsigned)1 << i)))
       return (i << shift) + base;
   }
 
@@ -2768,9 +3100,9 @@ XFORM_NONGCING static uintptr_t hamt_find_free_code(Scheme_Hash_Tree *tree, int
   minpos = mzHAMT_WORD_SIZE;
   for (i = mzHAMT_WORD_SIZE; i--; ) {
     if (!HASHTR_SUBTREEP(tree->els[i])) {
-      uintptr_t code = (i << shift) + base;
+      uintptr_t code = ((unsigned)i << shift) + base;
       if (_mzHAMT_CODE(tree, i, mzHAMT_WORD_SIZE) == code)
-        return code + (1 << (shift + mzHAMT_LOG_WORD_SIZE));
+        return code + ((unsigned)1 << (shift + mzHAMT_LOG_WORD_SIZE));
       else
         return code;
     } else {
@@ -2784,14 +3116,14 @@ XFORM_NONGCING static uintptr_t hamt_find_free_code(Scheme_Hash_Tree *tree, int
   }
 
   return hamt_find_free_code((Scheme_Hash_Tree *)tree->els[minpos],
-                             (minpos << shift) + base,
+                             ((unsigned)minpos << shift) + base,
                              shift + mzHAMT_LOG_WORD_SIZE);
 }
 
-static Scheme_Hash_Tree *make_hash_tree(int eql_kind, int val_kind, int popcount)
+static Scheme_Hash_Tree *make_hash_tree(int eql_kind, int popcount)
 {
   Scheme_Hash_Tree *ht;
-  int kind = val_kind | (eql_kind ? (HASHTR_HAS_CODE | HASHTR_HAS_VAL) : 0);
+  int kind = (eql_kind ? (HASHTR_HAS_CODE | HASHTR_HAS_VAL) : 0);
 
   ht = hamt_alloc(kind, popcount);
 
@@ -2807,22 +3139,17 @@ static Scheme_Hash_Tree *make_hash_tree(int eql_kind, int val_kind, int popcount
 
 Scheme_Hash_Tree *scheme_make_hash_tree(int eql_kind)
 {
-  return make_hash_tree(eql_kind, HASHTR_HAS_VAL, 0);
-}
-
-Scheme_Hash_Tree *scheme_make_hash_tree_set(int eql_kind)
-{
-  return make_hash_tree(eql_kind, 0, 0);
+  return make_hash_tree(eql_kind, 0);
 }
 
 Scheme_Hash_Tree *scheme_make_hash_tree_of_type(Scheme_Type stype)
 {
   if (stype == scheme_eq_hash_tree_type)
-    return scheme_make_hash_tree(0);
+    return scheme_make_hash_tree(SCHEME_hashtr_eq);
   else if (stype == scheme_hash_tree_type)
-    return scheme_make_hash_tree(1);
+    return scheme_make_hash_tree(SCHEME_hashtr_equal);
   else
-    return scheme_make_hash_tree(2);
+    return scheme_make_hash_tree(SCHEME_hashtr_eqv);
 }
 
 Scheme_Hash_Tree *scheme_make_hash_tree_placeholder(int eql_kind)
@@ -2832,12 +3159,12 @@ Scheme_Hash_Tree *scheme_make_hash_tree_placeholder(int eql_kind)
 {
   Scheme_Hash_Tree *ht, *sub;
 
-  ht = make_hash_tree(eql_kind, 0, 1);
+  ht = make_hash_tree(eql_kind, 1);
   ht->iso.so.type = scheme_hash_tree_indirection_type;
   ht->count = 0;
   ht->bitmap = 1;
 
-  sub = make_hash_tree(eql_kind, HASHTR_HAS_VAL, 0);
+  sub = make_hash_tree(eql_kind, 0);
   ht->els[0] = (Scheme_Object *)sub;
 
   return ht;
@@ -2860,6 +3187,7 @@ Scheme_Object *scheme_eq_hash_tree_get(Scheme_Hash_Tree *tree, Scheme_Object *ke
   int pos;
 
   h = PTR_TO_LONG((Scheme_Object *)key);
+  h = HAMT_REORDER(h);
 
   tree = hamt_assoc(resolve_placeholder(tree), h, &pos, 0);
   if (!tree)
@@ -2883,6 +3211,9 @@ Scheme_Object *scheme_hash_tree_get_w_key_wraps(Scheme_Hash_Tree *tree, Scheme_O
   int stype, pos;
 
   tree = resolve_placeholder(tree);
+  if (!tree->count)
+    return NULL;
+
   stype = SCHEME_TYPE(tree);
   
   if (stype == scheme_eq_hash_tree_type)
@@ -2893,6 +3224,7 @@ Scheme_Object *scheme_hash_tree_get_w_key_wraps(Scheme_Hash_Tree *tree, Scheme_O
     h = to_unsigned_hash(scheme_equal_hash_key(key));
   } else
     h = to_unsigned_hash(scheme_eqv_hash_key(key));
+  h = HAMT_REORDER(h);
 
   tree = hamt_assoc(tree, h, &pos, 0);
   if (!tree)
@@ -2939,6 +3271,7 @@ Scheme_Hash_Tree *scheme_hash_tree_set_w_key_wraps(Scheme_Hash_Tree *tree, Schem
     h = to_unsigned_hash(scheme_equal_hash_key(ekey));
   } else
     h = to_unsigned_hash(scheme_eqv_hash_key(key));
+  h = HAMT_REORDER(h);
   
   in_tree = hamt_assoc(resolve_placeholder(tree), h, &pos, 0);
   if (!in_tree) {
@@ -3019,7 +3352,7 @@ Scheme_Hash_Tree *scheme_hash_tree_set_w_key_wraps(Scheme_Hash_Tree *tree, Schem
         return tree;
       else {
         /* new hash collision */
-        in_tree = hamt_make2(SCHEME_HASHTR_KIND(in_tree) | HASHTR_HAS_CODE, 0,
+        in_tree = hamt_make2(SCHEME_HASHTR_KIND(in_tree) | HASHTR_HAS_VAL | HASHTR_HAS_CODE, 0,
                              0, in_tree->els[pos], mzHAMT_VAL(in_tree, pos),
                              1, key, val);
         in_tree->iso.so.type = scheme_hash_tree_collision_type;
@@ -3047,11 +3380,18 @@ static int hamt_equal_entries(int stype, void *eql_data,
         return SAME_OBJ(v1, v2);
     }
   } else if (stype == scheme_hash_tree_type) {
-    if (scheme_recur_equal(k1, k2, eql_data))
-      return scheme_recur_equal(v1, v2, eql_data);
+    if (eql_data) {
+      if (scheme_recur_equal(k1, k2, eql_data))
+        return scheme_recur_equal(v1, v2, eql_data);
+    } else
+      return scheme_equal(k1, k2);
   } else {
-    if (scheme_eqv(k1, k2))
-      return scheme_recur_equal(v1, v2, eql_data);
+    if (scheme_eqv(k1, k2)) {
+      if (eql_data)
+        return scheme_recur_equal(v1, v2, eql_data);
+      else
+        return 1;
+    }
   }
   return 0;
 }
@@ -3165,6 +3505,17 @@ int scheme_eq_hash_tree_subset_of(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2)
   return hamt_eq_subset_of(t1, t2, 0, scheme_eq_hash_tree_type, NULL);
 }
 
+int scheme_hash_tree_subset_of(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2)
+{
+  t1 = resolve_placeholder(t1);
+  t2 = resolve_placeholder(t2);
+
+  if (t1->count > t2->count)
+    return 0;
+  
+  return hamt_subset_of(t1, t2, 0, SCHEME_TYPE(t1), NULL);
+}
+
 int scheme_eq_hash_tree_subset_match_of(Scheme_Hash_Tree *t1, Scheme_Hash_Tree *t2)
 /* assumes that `t1` and `t2` are sets, as opposed to maps */
 {
